      SUBROUTINE FPRINT (numsubsfinal) 


!   ****************************    NOTICE   **************************** 
 
!    This "Image Quality of Fingerprint" (IQF) software was developed by 
!                            The MITRE Corporation
!    and was produced for the U.S. Government under Contract numbers 
!    W15P7T-05-C-F600 and W15P7T-07-C-F600 and is subject to the Rights in
!    Data General clause, at FAR 52.227-l4.
	
!   Redistribution and use in source and binary forms, with or without
!   modification, are permitted provided that the following conditions are met:
	
!    -Redistribution of source code must retain the above copyright notice,
!    this list of conditions, and the following disclaimer.
	
!    -Redistribution in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation and/or other materials provided with the distribution.
	
!    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
!    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
!    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
!    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
!    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
!    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
!    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
!    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
!    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
!    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
!    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
      
!   ********************************************************************* 
 
 
 
! This Subroutine:                                                                                                                                  
!  Finds the fingerprint in the image, then determines the locations of a set of square, same size windows   
!  (aka "subimages") which are within the found fingerprint area, and to which IQ computations 
!  are subsequently applied (outside of this subroutine).
!  This subroutine is primarily for inked or livescan fingerprints; latent fingerprints generally have too much
!  complex extraneous structure surrounding the latent print for this subroutine to work effectively.


! Basic Procedure:
!  - Fingerprint area is found within image, defined by left/right vertical edges & horizontal top/bot edges;
!    image must be 8bpp grayscale containing a single fingerprint with positive polarity (dark ridges & light vallies).
!  - A square window is defined whose width, Fprintwindow, is 60% (v1.0) of the width of the found fingerprint, 
!    with maximum & minimum size constraints defined in GlobalData
!  - This window travels throughout the fingerprint area, across and down, where sequential windows substantially overlap
!  - The upper left corner of each window is determined.
!  - Windows that include extremely dense/dark/opaque areas or very low-structure areas are weeded-out from further consideration.

!   Compute time is sensitive to the x,y shift increment used for window locations; too small a shift value
!   leads to inordinately long compute times (because it implies FFT on a large number of windows needs to be performed),
!   too large a shift value results in true best quality area being only partially measured.  x, y Shift value of 0.10 inches
!   is good compromise.  Compute time is also dependent on window width, see below comments on Brenner FFT.

!   Some of the major parameters:
!   Fprintwindow = pixel width of measurement window ("subimage"), declared Integer in Globaldata module
!   numsubsfinal (INTENT:OUT) = final number of 'good' candidate Fprintwindows, to which IQ is applied
!   ppi = resolution of digitized/scanned fingerprint image
!   xcoord(), ycoord() = upper left corner of window (subimage),  declared Integer array in Globaldata module
!   coordinate reference is 0,0: if ULcorner of subimage #1 coincides with ULcorner of entire image, then xcoord(1)=0, ycoord(1)=0  


      USE globaldata
	
      INTEGER :: colgap, Right, Rightedgecol, stripthickness, Vstripheight, Hstripwidth
      INTEGER :: blokwidth, blokheight, top, topedgerow, botedgerow
      INTEGER, ALLOCATABLE, DIMENSION(:) :: stripavg, ULcol, ULrow 
      REAL, ALLOCATABLE, DIMENSION(:) :: blokavg, blokstandev

	

! SetUp Constants & initial values	
	increment = ppi * shiftpixels / 500	! subimage (square window) shift increment in x & y in pixels
	maxheight = ppi * maxheightpixels / 500		! clamps max Fprint height to 1.4" (see "botedgerow")
	stripthickness = ppi * 2 / 500	! pixel width of vertical or horizontal strip (for fprint edge finder)
	IF(stripthickness < 1) stripthickness = 1
	colgap = ppi * 20 / 500		! add window if there is uncovered area in Fprint > colgap pixels wide
	adjustment = .7	  ! used in computing "itest", which is a test value to ascertain left/right fprint edges
    	DarkThresholdOne = 20. 	! for "very dense" test using avg gray of little square bloks	
	StanDevThreshold = .3   ! for "very low structure" test using standev gray oflittle square bloks
	DarkThresholdTwo = 80.	! for "very low structure" test using avg gray of little square bloks
	blokwidth = .08 * REAL(ppi)	! .08" little block width & height (for dense & low structure tests)
	blokheight = blokwidth
	bloksamples = REAL(blokwidth * blokheight)
      	MinWindow = MinWindowWidth * ppi / 500
      	MaxWindow = MaxWindowWidth * ppi / 500
	
      	IF(MinWindow > totwidth .OR. MinWindow > totheight) THEN	! image is too small for specified windows, so skip it
   		numsubsfinal = 0
   		RETURN
   	ENDIF

      	IF(totheight >= ppi) THEN	! go-in from each image edge for calculations to find fprint edges; mostly used to avoid
      		margin = ppi/20		! potential inked fprint block lines or labeling near edges, while not adversely
      	ELSE				! affecting livescan search (which usually doesnt have such obstructions)
      		margin = ppi/40		! margin=0.05" if image height > 1", margin=0.025" for smaller images
      	ENDIF
    
 ! Vertical narrow strips are defined across image width, which are used to locate left/right fprint edges,  
 ! Avg of each vertical strip is computed:
 
 	i = totheight / stripthickness;     j = totwidth / stripthickness
 	k = MAX(i,j) + 2
      	ALLOCATE(stripavg(k))	
      	stripavg = 0     	  
      	Vstripheight = totheight - 2*margin
      	numVstrips = (totwidth - 2*margin) / stripthickness
      	Vsamples = REAL(stripthickness * Vstripheight)
   	DO j = 1, numVstrips
		isum = 0
		DO k = 1, Vstripheight
			DO m = 1, stripthickness
			   isum = isum + image((margin+(j-1)*stripthickness+m-1), (margin + k - 1))  ! image(col,row), origin at 1,1
			END DO 
		END DO
		stripavg(j) = NINT(REAL(isum) / Vsamples)	! avg gray of jth vertical strip, going left to right
   	END DO	
   	maxavgrayV = MAXVAL(stripavg(1:numVstrips),1)
 	minavgrayV = MINVAL(stripavg(1:numVstrips),1)
 
  
! **************************** FIND LEFT & RIGHT EDGES OF FINGERPRINT: ********************************************  
! there are at least 5 possibilities:  
!  	(1) left & right Fprint edges exist within image width  
!	(2) left Fprint edge coincides with left image edge (right edge exists)
!	(3) right Fprint edge coincides with right image edge (left edge exists)  
!	(4) left & right Fprint edges conincide with left & right image edges
!	(5) no Fprint, or multiple Fprints in image: L/R edges may be found but are meaningless
!  L/R edges are found from assessment of avg gray of narrow vertical strips

  	! find Leftside edge:
  	itest = NINT(adjustment * REAL(maxavgrayV-minavgrayV)) + minavgrayV  
  	left = 1;     right = numVstrips	! initial left/right fprint edges in terms of Vstrip number
  	DO j= 1, numVstrips	
 		IF(stripavg(j) < itest) THEN
 			Left = j
 			EXIT 
 		ENDIF
 	END DO
  	leftedgecol = stripthickness * Left - stripthickness + 1
! 	PRINT*, 'prelim leftedge= ',leftedgecol

 	IF(leftedgecol > (totwidth - MinWindow/2)) THEN       ! use MinWindow to cover case of small fprint
 		leftedgecol = margin
 		rightedgecol = totwidth - margin
 		Fprintwindow = MinWindow     ! dont know fprint width so use min window width
 		GOTO 10
 	ELSEIF((leftedgecol > (totwidth - MinWindow)) .AND. (leftedgecol <= (totwidth - MinWindow/2))) THEN
 		leftedgecol = totwidth - MinWindow - margin
 		rightedgecol = totwidth - margin
 		Fprintwindow = MinWindow  ! initial fprint width is close to MinWindow width, so use MinWindow
 		GOTO 10
	ENDIF


	ijk = MinWindow / (2*stripthickness)	! find RIGHTSIDE EDGE strip
 	DO j= left+ijk, numVstrips	
 		IF(stripavg(j) > itest) THEN
 			Right = j
 			EXIT
 		ENDIF
 	END DO	
   	rightedgecol = stripthickness * Right	
! 	PRINT*, 'prelim rightedgecol   ', rightedgecol
   	
   	IF((leftedgecol > rightedgecol) .OR. ((rightedgecol-leftedgecol) < MinWindow)) THEN	 
   		leftedgecol = margin		! L/R edges not found, so use entire image width minus margins
   		rightedgecol = totwidth - margin		
		Fprintwindow = MinWindow    ! dont know fprint width so use min window width
		GOTO 10
   	ENDIF
 !***********************************************************************************************************************



! Define width of square window in pixels = Fprintwindow = subimage width = window to which FFT will be applied  (for each location) 

	Fprintwindow = NINT(widthfactor * (rightedgecol - leftedgecol))	 ! Fprintwindow = widthfactor x fingerprintwidth

10	IF(Fprintwindow < MinWindow) THEN
		Fprintwindow = MinWindow
	ELSEIF(Fprintwindow > MaxWindow) THEN
		Fprintwindow = MaxWindow
	ENDIF
	
	IF(Fprintwindow > (totwidth-2*margin)) THEN
		Fprintwindow = totwidth - 2*margin - 1
		leftedgecol = margin
		rightedgecol = totwidth - margin
	ENDIF
	
	IF(MOD(Fprintwindow,2) /= 0) Fprintwindow = Fprintwindow + 1  ! force width to even number for faster FFT compute
	
!	Brenner FFT compute time can be alot longer for certain widths, generally those that have very few prime factors
!	(especially few factors of 2); experimentation was run to identify the true 'bad' widths, in range: 200 to 1200 pixels
!	The following checks Fprintwindow against these bad widths and, if a match is found, Fprintwindow is adjusted down or
!	up a few even pixels to change it to a 'good', short compute time value.
!	This is not applied to Latent image, which has user predefined size.
	n = 0
	DO i = 1, 186	! number of NGwindows defined in Globaldata (minus the 5 leading & trailing "1s")
		IF(Fprintwindow == NGwindow(i)) THEN
			n = i
			EXIT
		ENDIF
	END DO
		
	IF(n /= 0) THEN
		DO j = 0, 4	! enough range for up to & incl. 3 sequentially adjacent NGwindows on each side of Fprintwindow
			IF((Fprintwindow - (j*2)) /= NGwindow(n-j)) THEN 
				Fprintwindow = Fprintwindow - (j*2) 
				EXIT
			ENDIF
			IF((Fprintwindow + (j*2)) /= NGwindow(n+j)) THEN
				Fprintwindow = Fprintwindow + (j*2)
				EXIT	
			ENDIF
		END DO
	ENDIF

   	IF(Fprintwindow > totwidth .OR. FprintWindow > totheight) THEN
   		numsubsfinal = 0
   		DEALLOCATE(stripavg)
   		RETURN
   	ENDIF


	IF(MinWindow == MaxWindow) Fprintwindow = MinWindow   ! if min=max, a constant window width was specified in gloabaldata 
	
	

! **************************** FIND TOP & BOTTOM EDGES OF FINGERPRINT: ****************************************
! Final Fprintwindow has already been determined; top/bottom edges are only used to decrease IQ search area (minimize # of windows)

! Compute average gray of each horizontal strip:
       	stripavg = 0
       	Hstripwidth = totwidth - 2*margin
      	numHstrips = (totheight - 2*margin) / stripthickness
      	Hsamples = REAL(stripthickness * Hstripwidth)
   	DO j = 1, numHstrips
		isum = 0
		DO k = 1, Hstripwidth
			DO m = 1, stripthickness	
				isum = isum + image((margin-1+k), (margin+stripthickness*(j-1)+m-1))  ! image(col,row), origin at 1,1
			END DO 
		END DO
		stripavg(j) = NINT(REAL(isum) / Hsamples)	! avg gray of jth vertical strip, going left to right
!   		print*, '       ',j, stripavg(j)
   	END DO	
   	maxavgrayH = MAXVAL(stripavg(1:numHstrips),1)
 	minavgrayH = MINVAL(stripavg(1:numHstrips),1)
  	
  	
  	itest = NINT(adjustment * REAL(maxavgrayH-minavgrayH)) + minavgrayH   
  	top = 1
  	DO j= 1, numHstrips	
 		IF(stripavg(j) < itest) THEN
 			top = j
 			EXIT
 		ENDIF
 	END DO
 	topedgerow = margin + stripthickness * top - stripthickness + 1
 	IF(((totheight-2*margin) - topedgerow) < Fprintwindow) topedgerow = margin
 	
! 	clamp max usable height of fprint to "maxheight", this is useful in some cases: it can
! 	potentially avoid labeling at image bottom, or crease/flexure area, or alot of empty space
!	at bottom of image with abnormally high image height 
 	botedgerow = totheight - margin	
 	IF((totheight - margin - topedgerow) > maxheight) botedgerow = topedgerow + maxheight - 1
 	
 !**********************************************************************************************************
 
 

!  PRINT*, 'final L/R/T/B edges:  ',leftedgecol,rightedgecol,topedgerow,botedgerow		
							  
   
! Determine number of measurement windows (subimages) across and down, within found Fprint area:    
       numacrosswins = (((rightedgecol - leftedgecol) - Fprintwindow) / increment) + 1
       iflagx = 0
       j = (rightedgecol - leftedgecol) - increment*(numacrosswins-1) + Fprintwindow
       IF(j > colgap) THEN 	! if rightedge of rightmost window is > colgap columns from rightedgecol 
        	iflagx = 1	! then add another window whose right edge is at rightedgecol
      		numacrosswins = numacrosswins + 1
      	ENDIF
      
       numdownwins = (((botedgerow-topedgerow) - Fprintwindow) / increment) + 1
       iflagy = 0
       j = botedgerow - (increment*(numdownwins-1) + Fprintwindow)
       IF(j > colgap) THEN
       		iflagy =1 
 		numdownwins = numdownwins + 1
 	ENDIF
 	numsubsfinal = numacrosswins * numdownwins	! initial total number of windows, assuming all are good to use


! **** WEED-OUT 'POOR' MEASUREMENT WINDOWS **********************************************
! Grid the Fingerprint Area into small Square Blocks & Compute Avg Gray & StanDevGray in Each Block.
! Purpose is to weed-out windows that are very dense or have very low structure content.
!   Weed-Out window If AvgGray of too many little blocks < global DarkThreshold (this is a 'very dense' block).
!   Weed-Out window If StanDevGray of too many little blocks < 30% of StanDevGray of entire Fprint & little block is dense
!   (this is a 'very low structure' block).
!  "little block" is square block whose width is much smaller than the square Fprintwindow
	
	numacrossbloks = (rightedgecol - leftedgecol) / blokwidth
	numdownbloks = (botedgerow - topedgerow) / blokheight
	nad = numacrossbloks * numdownbloks

! Allocate memory for total number of adjacent, sequential little blocks within entire fprint area:  
       ALLOCATE(ULcol(nad), ULrow(nad))
       ALLOCATE(blokstandev(nad), blokavg(nad)) 

!	Compute standev & avg value of Entire Fprint area:
	ksum = 0;     xsum = 0.0    
	xxx = REAL((rightedgecol-leftedgecol+1) * (botedgerow-topedgerow+1)) 
	DO irow= topedgerow, botedgerow 	
		DO icol = leftedgecol, rightedgecol   
			ipixel = image(icol, irow) 
			ksum = ksum + ipixel
			xsum = xsum + REAL(ipixel * ipixel)
		END DO
	END DO
	rksum = REAL(ksum)
	avgFprint = rksum / xxx 
	varFprint = (xxx * xsum - rksum * rksum) / (xxx * (xxx - 1.))
	IF(varFprint < 0.) varFprint = .0001	! dont need dble precision accuracy but must avoid neg value
	StanDevFprint = SQRT(varFprint)

! Compute avg & standev of each little block within fprint area:	
	nn = 0			! counter for all blocks
! 	PRINT*, 'blokwidth, numacross, numdown ', blokwidth, numacrossbloks, numdownbloks
	DO j = 1, numdownbloks
		DO i = 1, numacrossbloks
		    ksum = 0;     nsum = 0
		    nn = nn + 1
		    DO k = 1, blokheight
			DO m = 1, blokwidth						
			     mpixel = image(((i-1)*blokwidth+m+leftedgecol-1), ((j-1)*blokheight+k+topedgerow-1))  
			     ksum = ksum + mpixel
			     nsum = nsum + mpixel * mpixel
			END DO 
		   END DO
		   rksum = REAL(ksum)
		   blokavg(nn) = rksum / bloksamples		! avg gray of block
		   blokvar = (bloksamples * REAL(nsum) - rksum * rksum) / (bloksamples * (bloksamples - 1.))
		   IF(blokvar < 0.) blokvar = .0001	
		   blokstandev(nn) = SQRT(blokvar)
		   ULcol(nn) = (i-1)*blokwidth + leftedgecol	! UpperLeft corner col,row of block
		   ULrow(nn) = (j-1)*blokheight + topedgerow        
	        END DO
	END DO

!	at this point,  nad = nn = total number of little blocks in entire fprint area



	
! Define ULcorner of each measurement window ("subimage") and delete any window that 
! contains too many little dense blocks or too many little low structure blocks.
! Coord reference: if first subimage is upperleft corner of entire image, the ULcorner coord of that subimage is 0,0 not 1,1
	n = 0	! index counter for window 
	m = 0	! counts number of windows weeded-out
	fpwfpw = REAL(Fprintwindow * Fprintwindow)	! # of samples in square window

!	Now perform window weeding-out process	
	DO j = 1, numdownwins
		DO i = 1, numacrosswins
			n = n+1
			IF(iflagx==1 .and. i==numacrosswins) THEN
				xcoord(n) = rightedgecol - Fprintwindow
			ELSE
				xcoord(n) = leftedgecol + i*increment-increment - 1	
			ENDIF
	
			IF(iflagy==1 .and. j==numdownwins) THEN
				ycoord(n) = botedgerow - Fprintwindow
			ELSE
				ycoord(n) = topedgerow + j*increment-increment - 1      
			ENDIF						
			
!			Determine Number of Very-Dense Little Blocks in Window Area:			
			iflag = 0;   numbadAvg = 0;    numbadSD = 0
			
			DO k = 1, nad	! nad=number of little blocks in Entire fprint area
			     ck = ULcol(k) + blokwidth/4	! shrink blok to halfsize centered in orig blok
			     rk = ULrow(k) + blokheight/4	! then see if any portion of shrunk blok is in window
			     IF(blokavg(k) < DarkThresholdOne) THEN
			         IF((ck < (xcoord(n)+Fprintwindow)) .AND. ((ck+blokwidth/2) > xcoord(n)) &
				    .AND. ((rk+blokheight/2) > ycoord(n)) .AND. (rk < (ycoord(n)+Fprintwindow))) THEN
						numbadAvg = numbadAvg + 1
			         ENDIF 
			     ENDIF
			END DO
			IF((REAL(numbadAvg) * bloksamples / fpwfpw) > .03)  iflag = 1    ! flag if dark area > 3% of window area
									    	  ! 3% sounds very low but higher % accepts windows
									    	  ! with too much visually dense areas

!			Determine Number of Very-Low-Structure Little Blocks in Window Area;
!			Test is a combination of low standev and dark avg gray, which was found to work better than
!			using just standev.  On the other hand, if e.g. it is a totally blank white area (light gray)
!			then test is not met (no "numbadSD" found) but that is OK because such an area will have very 
!			low computed IQ and will not be selected as peak IQ; whereas a very dense area may have a high
!                       computed IQ in initial stage, prior to final IQ normalizations.						
			IF(iflag ==  0) THEN
				DO k = 1, nad
			     	   ck = ULcol(k) + blokwidth/4
			     	   rk = ULrow(k) + blokheight/4
				   IF((blokstandev(k) / StanDevFprint) < StanDevThreshold .AND. blokavg(k) < DarkThresholdTwo) THEN
			         	IF((ck < (xcoord(n)+Fprintwindow)) .AND. ((ck+blokwidth/2) > xcoord(n)) &
				    	   .AND. ((rk+blokheight/2) > ycoord(n)) .AND. (rk < (ycoord(n)+Fprintwindow))) THEN
						  numbadSD = numbadSD + 1
			         	ENDIF
			           ENDIF 
			  	END DO
			 ENDIF
			 IF(numbadSD > 1)  iflag = 1   ! to skip low-structure test, use: IF(numbadSD > 99999) 
			IF(iflag == 1) THEN
				n = n-1	  ! skip the flagged window that has a dark patch(s) that cover too much of the window area
				m = m+ 1  ! counts number of weeded-out windows
			END IF	
		END DO	! numacrosswins
	END DO	! numdownwins

! **** END OF WINDOWS WEED-OUT PROCESS **********************************************


	DEALLOCATE(ULcol, ULrow, stripavg) 
	DEALLOCATE(blokstandev, blokavg)                                                                                                                       

	numsubsfinal = numsubsfinal - m		! final number of good windows for which IQ is computed

	
	IF(numsubsfinal < 1) THEN  ! no good windows found, so define one window for IQ: L/R centered, biased towards top
		numsubsfinal = 1   ! this is the best that can be done for this case
		xcoord(1) = leftedgecol + (rightedgecol - leftedgecol - Fprintwindow) / 2
		ycoord(1) = topedgerow + (botedgerow - topedgerow - Fprintwindow) / 3
!		PRINT*, 'Caution, this Fprint may be extremely dense or light or has no-structure    ', imgID
	ENDIF
	
!	PRINT*, 'initial,final numsubsfinal:  ', (numsubsfinal+m), numsubsfinal
	
        
      END SUBROUTINE FPRINT   
      
   
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      
      
	  
