      SUBROUTINE computeIQ (prelim_IQ, contrast, avgGray) 
                                                                                                                                  
!   ****************************    NOTICE   **************************** 
 
!    This "Image Quality of Fingerprint" (IQF) software was developed by 
!                            The MITRE Corporation
!    and was produced for the U.S. Government under Contract numbers 
!    W15P7T-05-C-F600 and W15P7T-07-C-F600 and is subject to the Rights in
!    Data General clause, at FAR 52.227-l4.
	
!   Redistribution and use in source and binary forms, with or without
!   modification, are permitted provided that the following conditions are met:
	
!    -Redistribution of source code must retain the above copyright notice,
!    this list of conditions, and the following disclaimer.
	
!    -Redistribution in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation and/or other materials provided with the distribution.
	
!    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
!    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
!    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
!    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
!    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
!    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
!    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
!    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
!    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
!    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
!    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
      
!   ********************************************************************* 

! This subroutine computes the preliminary IQ value from the given subimage
! the power spectrum is normalized by total power, then filtered by HVS filter                                                                                                         
                                                                                                                                                                                                                             
!  HVS filter model, H(r), is the square of a rotationally symmetric MTF representing the HVS:                                                                                                                                   
!  H(r) = ((.2 + .45r)*exp(-.18r))                                                                                                                                                                                                       
!             r = number of cycles/degree subtended by the eye:                                                                
!             r = .0174536(viewdist/spot)sqrt[(i/xlength)**2 + (j/ylength)**2]                                                       
!             viewdist = pseudo, nominal distance (mm): eye-to-display                                                                 
!             spot = pseudo, nominal display pixel width (mm)                                                                              
!	      HVS peak location along power spectrum cycles/pixelwidth axis:
!	      (292.774*spot) / viewdist = cy/pixewidth of HVS peak
!	      spot=0.6, viewdist=351.3288   sets peak at 0.50 cy/pixwidth
!             HVSparm  = "T" in Opt.Eng., April'92 paper = (2*viewdist) Tan(0.5 degrees) / spot 


      USE GLOBALDATA
	      
      DOUBLE PRECISION :: dbl_iq, dcpower, acpower, totpower
      CHARACTER(LEN=3) :: oddevenwidth                                                                                                                                                                                                                                                                
       
 	  
      dcpower = DBLE(power (0, 0)) 
      acpower = 0.D0	  
      contrast = 0.0
      rxlength = REAL(xlength) 		! image width, height in pixels
      rylength = REAL(ylength) 	
      dbl_iq = 0.0D0
     
                                                                                                                                   
!     Compute Total Power (AC+DC Power):                                                                                                           	  
      totpower = SUM(DBLE(power(0, -ylength/2 : ylength/2)))	! along (y) vertical axis at x=0
      DO i = 1, xlength / 2 
           DO  j = - ylength / 2, ylength / 2   
                  totpower = totpower + 2.D0 * DBLE(power(i, j))   
	   END DO 
      END DO 
      
      acpower = totpower - dcpower
      avgGray = SNGL(DSQRT(dcpower)) / (rxlength * rylength) 
                                                                                                                          
!     Compute Contrast:                                                                                                           
	  IF (dcpower < TinyDble) THEN
		contrast = 0.0
	  ELSE
	  	contrast = SNGL(DSQRT(acpower / dcpower))	
	  ENDIF
	  
		
!   Normalize by total power, which is, overall, best normalization to find peakprelim_IQ window 
!   from set of candidate windows (in iqf subroutine); the peakprelim_IQ is adjusted later 
!   for contrast & avg gray to get final peakIQ that is intercomparable across different fprint images.

    IF(totpower < TinyDble) THEN
    	power = 0.D0
    ELSE	
    	 power = power / SNGL(totpower)
    ENDIF
    
 ! Compute HVS filter & multiply by image power values:
      DO  i = 0, xlength/2 
           rxl = (i / rxlength) **2 
           DO  j = -ylength/2, ylength/2 
	   		  ryl = (j / rylength)**2
	   		  radialfreq = SQRT (rxl + ryl) 
			  cyperdeg = HVSparm * radialfreq 
              		  hvsmtf2 = ((0.2 + 0.45 * cyperdeg) * EXP(- 0.18 * cyperdeg)) **2 
			  power(i,j) = power(i,j) * hvsmtf2
	   END DO
      END DO


! Compute prelim_IQ value from normalized, filtered power spectrum:
! (take advantage of bilateral symmetrism property of real image transform)
     
      IF(MOD(xlength,2) == 0) THEN		
      		oddevenwidth = 'evn'	! xlength is even so exclude leftside column at -xlength/2
      ELSE
      		oddevenwidth = 'odd'	! xlength is odd so include leftside column at -xlength/2
      ENDIF
      
      ihalfx = xlength/2  
      DO  i = 0, xlength/2 
           rxl = (i / rxlength)**2 
           DO  j = -ylength/2, ylength/2 
	   	  ryl = (j / rylength)**2
	   	  radialfreq = SQRT (rxl + ryl)      ! IQ summation: cycles per pixelwidth = .01 to 1/sqrt(2)
		  IF (radialfreq >= .01 .AND. radialfreq <= .707107) THEN
			 IF((i==0) .OR. (i==ihalfx .AND. oddevenwidth=='evn')) THEN 
				   	dbl_iq = dbl_iq + DBLE(power(i,j)) 
		     	 ELSE
		     	  		dbl_iq = dbl_iq + 2.D0 * DBLE(power(i,j)) 
			 ENDIF
		  ENDIF
	   END DO
      END DO
 

        prelim_IQ =  SNGL(dbl_iq)
     
     
      END SUBROUTINE computeIQ   










                                                     
